use crate::parser::ParseData;
use proc_macro2::TokenStream;
use quote::quote;
use syn::{LitStr, Type};

impl ParseData {
	pub(super) fn gen_schema(self) -> syn::Result<TokenStream> {
		match self {
			Self::Struct(fields) => gen_struct(fields),
			Self::Unit => gen_unit(),
			_ => unimplemented!()
		}
	}
}

fn gen_struct(fields: Vec<(LitStr, Type)>) -> syn::Result<TokenStream> {
	let field_name = fields.iter().map(|(name, _)| name);
	let field_ty = fields.iter().map(|(_, ty)| ty);

	let openapi = path!(::openapi_type::openapi);
	Ok(quote! {
		{
			let mut properties = <::openapi_type::indexmap::IndexMap<
				::std::string::String,
				#openapi::ReferenceOr<::std::boxed::Box<#openapi::Schema>>
			>>::new();
			let mut required = <::std::vec::Vec<::std::string::String>>::new();

			#({
					const FIELD_NAME: &::core::primitive::str = #field_name;
					let mut field_schema = <#field_ty as ::openapi_type::OpenapiType>::schema();
					add_dependencies(&mut field_schema.dependencies);

					// fields in OpenAPI are nullable by default
					match field_schema.nullable {
						true => field_schema.nullable = false,
						false => required.push(::std::string::String::from(FIELD_NAME))
					};

					match field_schema.name.as_ref() {
						// include the field schema as reference
						::std::option::Option::Some(schema_name) => {
							let mut reference = ::std::string::String::from("#/components/schemas/");
							reference.push_str(schema_name);
							properties.insert(
								::std::string::String::from(FIELD_NAME),
								#openapi::ReferenceOr::Reference { reference }
							);
							dependencies.insert(
								::std::string::String::from(schema_name),
								field_schema
							);
						},
						// inline the field schema
						::std::option::Option::None => {
							properties.insert(
								::std::string::String::from(FIELD_NAME),
								#openapi::ReferenceOr::Item(
									::std::boxed::Box::new(
										field_schema.into_schema()
									)
								)
							);
						}
					}
			})*

			#openapi::SchemaKind::Type(
				#openapi::Type::Object(
					#openapi::ObjectType {
						properties,
						required,
						.. ::std::default::Default::default()
					}
				)
			)
		}
	})
}

fn gen_unit() -> syn::Result<TokenStream> {
	let openapi = path!(::openapi_type::openapi);
	Ok(quote! {
		#openapi::SchemaKind::Type(
			#openapi::Type::Object(
				#openapi::ObjectType {
					additional_properties: ::std::option::Option::Some(
						#openapi::AdditionalProperties::Any(false)
					),
					.. ::std::default::Default::default()
				}
			)
		)
	})
}
