#![warn(missing_debug_implementations, rust_2018_idioms)]
#![deny(broken_intra_doc_links)]
#![forbid(unsafe_code)]
//! This crate defines the macros for `#[derive(OpenapiType)]`.

use proc_macro::TokenStream;
use proc_macro2::TokenStream as TokenStream2;
use quote::quote;
use syn::{parse_macro_input, Data, DeriveInput, LitStr, TraitBound, TraitBoundModifier, TypeParamBound};

#[macro_use]
mod util;
//use util::*;

mod codegen;
mod parser;
use parser::*;

/// The derive macro for [OpenapiType][openapi_type::OpenapiType].
#[proc_macro_derive(OpenapiType)]
pub fn derive_openapi_type(input: TokenStream) -> TokenStream {
	let input = parse_macro_input!(input);
	expand_openapi_type(input).unwrap_or_else(|err| err.to_compile_error()).into()
}

fn expand_openapi_type(mut input: DeriveInput) -> syn::Result<TokenStream2> {
	let ident = &input.ident;
	let name = ident.to_string();
	let name = LitStr::new(&name, ident.span());

	// prepare the generics - all impl generics will get `OpenapiType` requirement
	let (impl_generics, ty_generics, where_clause) = {
		let generics = &mut input.generics;
		generics.type_params_mut().for_each(|param| {
			param.colon_token.get_or_insert_with(Default::default);
			param.bounds.push(TypeParamBound::Trait(TraitBound {
				paren_token: None,
				modifier: TraitBoundModifier::None,
				lifetimes: None,
				path: path!(::openapi_type::OpenapiType)
			}))
		});
		generics.split_for_impl()
	};

	// parse the input data
	let parsed = match &input.data {
		Data::Struct(strukt) => parse_struct(strukt)?,
		Data::Enum(inum) => parse_enum(inum)?,
		Data::Union(union) => parse_union(union)?
	};

	// run the codegen
	let schema_code = parsed.gen_schema()?;

	// put the code together
	Ok(quote! {
		#[allow(unused_mut)]
		impl #impl_generics ::openapi_type::OpenapiType for #ident #ty_generics #where_clause {
			fn schema() -> ::openapi_type::OpenapiSchema {
				// prepare the dependencies
				let mut dependencies = <::openapi_type::indexmap::IndexMap<
					::std::string::String,
					::openapi_type::OpenapiSchema
				>>::new();

				// this function can be used to include dependencies of dependencies
				let add_dependencies = |deps: &mut ::openapi_type::indexmap::IndexMap<
					::std::string::String,
					::openapi_type::OpenapiSchema
				>| {
					while let ::std::option::Option::Some((dep_name, dep_schema)) = deps.pop() {
						if !dependencies.contains_key(&dep_name) {
							dependencies.insert(dep_name, dep_schema);
						}
					}
				};

				// create the schema
				let schema = #schema_code;

				// return everything
				const NAME: &::core::primitive::str = #name;
				::openapi_type::OpenapiSchema {
					name: ::std::option::Option::Some(::std::string::String::from(NAME)),
					nullable: false,
					schema,
					dependencies
				}
			}
		}
	})
}
